import uuid
import traceback
from pathlib import Path
from flask import redirect, send_file
from lyrebird import application
from lyrebird.version import VERSION
from flask_restful import Resource, request
from urllib.parse import urlencode
from lyrebird.mock import context
from lyrebird.log import get_logger


logger = get_logger()


class SnapshotImport(Resource):
    def get(self):
        queries = request.args
        new_query = {}
        application.active_menu = {
            'name': 'datamanager',
            'title': 'DataManager',
            'type': 'router',
            'path': '/datamanager'
        }

        group_id = queries.get('id')
        node = context.application.data_manager.id_map.get(group_id)
        if group_id and node:
            new_query['groupId'] = group_id

        elif queries.get('path'):
            if group_id:
                new_query['infoMsg'] = f'Snapshot {group_id} not found, use path argument instead'

            import_uri_link = queries['path']
            parent_path = queries.get('parent') or '/'
            parent_id = context.application.data_manager.add_group_by_path(parent_path)

            # is advanced save
            if queries.get('isAdvancedSave') == 'true':
                import_uri_uuid = str(uuid.uuid4())
                new_query['snapshotId'] = import_uri_uuid
                context.application.data_manager.snapshot_import_cache[import_uri_uuid] = {'link': import_uri_link}
                return redirect(f'/ui/?v={VERSION}#/datamanager/import?{urlencode(new_query)}')

            path = context.application.data_manager.read_snapshot_from_link(import_uri_link)

            try:
                group_id = context.application.data_manager.import_from_file(parent_id, path)
                new_query['groupId'] = group_id
            except Exception:
                new_query['errorMsg'] = f'Import snapshot error!'
                logger.error(f'Import snapshot error!\n {traceback.format_exc()}')
                return redirect(f'/ui/?v={VERSION}#/datamanager?{urlencode(new_query)}')

        if queries.get('isAutoActive') == 'true':
            context.application.data_manager.deactivate()
            context.application.data_manager.activate(group_id)

        display_info = queries.get('displayKey', '')
        if display_info:
            new_query['displayKey'] = display_info

        return redirect(f'/ui/?v={VERSION}#/datamanager?{urlencode(new_query)}')

    def post(self):
        if request.files:
            stream = request.files['file']
            if not stream:
                return application.make_fail_response('Missing required file')

            parent_id = request.form.get('parent_id') if request.form else None
            if not parent_id:
                return application.make_fail_response('Missing required argument: parent_id')

            filename = stream.filename or str(uuid.uuid4())
            path = context.application.data_manager.snapshot_workspace / filename
            stream.save(str(path))

            group_id = context.application.data_manager.import_from_file(parent_id, path)
            return application.make_ok_response(group_id=group_id)

        elif request.json:
            parent_id = request.json.get('parentId')
            name = request.json.get('snapshotName', '')
            if not parent_id:
                return application.make_fail_response('Missing required argument: parent_id')

            snapshot_id = request.json.get('snapshotId')
            snapshot_info = context.application.data_manager.snapshot_import_cache.get(snapshot_id)
            if not snapshot_info:
                return application.make_fail_response(f'Snapshot cache {snapshot_id} not found!')

            path = snapshot_info.get('path')
            if not Path(path).exists():
                link = snapshot_info.get('link')
                path = context.application.data_manager.read_snapshot_from_link(link)
                context.application.data_manager.snapshot_import_cache[snapshot_id]['path'] = path

            group_id = context.application.data_manager.import_from_file(parent_id, path, name=name)
            return application.make_ok_response(group_id=group_id)

        return application.make_fail_response('No import snapshot info found!')

    def put(self):
        if request.json:
            group_id = context.application.data_manager.import_from_local(request.json)
            return application.make_ok_response(group_id=group_id)

class SnapshotExport(Resource):
    def get(self, group_id):
        file = context.application.data_manager.export_from_remote(group_id)
        return send_file(file)

    def post(self):
        if not request.json.get('snapshot'):
            return application.make_fail_response('Missing required argument: snapshot')

        if not request.json.get('events'):
            return application.make_fail_response('Missing required argument: events')

        group_id, file_ = context.application.data_manager.export_from_local(request.json)
        req = send_file(file_)
        req.headers.add('SnapshotId', group_id)
        return req


class Snapshot(Resource):
    def get(self, snapshot_id):
        link = context.application.data_manager.snapshot_import_cache.get(snapshot_id, {}).get('link')
        if not link:
            application.make_fail_response(f'No import snapshot {snapshot_id} link found!')
        path = context.application.data_manager.read_snapshot_from_link(link)
        context.application.data_manager.snapshot_import_cache[snapshot_id]['path'] = path
        detail, output_path = context.application.data_manager.get_snapshot_file_detail(path)

        context.application.data_manager._remove_file([output_path])

        detail.pop('children')
        return application.make_ok_response(data=detail)



from lyrebird.config import ConfigException
from flask_restful import Resource
from lyrebird.mock import context
from flask import request, jsonify
from lyrebird import application
from lyrebird import log

logger = log.get_logger()


class Conf(Resource):
    """
    Lyrebird 及 插件 配置文件获取和修改
    """
    def get(self):
        return jsonify(application.config.raw())

    def put(self):
        try:
            context.application.conf = request.get_json()
            context.application.save()
            return context.make_ok_response()
        except Exception as e:
            return context.make_fail_response(str(e))

    def patch(self):

        update_conf = request.get_json()
        if update_conf is None or not isinstance(update_conf, dict):
            return application.make_fail_response('Request body must be a JSONObject!')

        if not update_conf:
            logger.warning('This request cannot modify config, request body is empty.')
            return application.make_ok_response()

        try:
            application._cm.override_config_field(update_conf)
            return application.make_ok_response()
        except ConfigException as e:
            return application.make_fail_response(str(e))



class TextHandler:

    @staticmethod
    def origin2flow(request_data):
        _data = request_data.decode('utf-8')
        return _data

    @staticmethod
    def flow2origin(flow_data):
        _data = flow_data.encode()
        return _data



from flask import Blueprint, send_file
from lyrebird import log
from pathlib import Path


CLIENT_ROOT_DIR = Path(__file__).parent/'../../client/static'


ui = Blueprint('ui', __name__, url_prefix='/ui', static_folder=str(CLIENT_ROOT_DIR))


logger = log.get_logger()

@ui.route('/')
def index():
    return send_file(str(CLIENT_ROOT_DIR/'index.html'))



import imp
import codecs
import shutil
from pathlib import Path
import lyrebird
from lyrebird import log
from lyrebird import application
from lyrebird.mock import context
from lyrebird.base_server import ThreadServer
from lyrebird.checker.kill import kill_flow

from .event import event
from .on_request import on_request
from .on_request_upstream import on_request_upstream
from .on_response import on_response
from .on_response_upstream import on_response_upstream
from .encoder import encoder
from .decoder import decoder


logger = log.get_logger()

scripts_tmp_storage = {}

TYPE_EVENT = 'event'
TYPE_ON_REQUEST = 'on_request'
TYPE_ON_REQUEST_UPSTREAM = 'on_request_upstream'
TYPE_ON_RESPONSE = 'on_response'
TYPE_ON_RESPONSE_UPSTREAM = 'on_response_upstream'
TYPE_ENCODER = 'encoder'
TYPE_DECODER = 'decoder'

FUNC_MAP_HANDLERS = {
    TYPE_EVENT: event,
    TYPE_ON_REQUEST: on_request,
    TYPE_ON_REQUEST_UPSTREAM: on_request_upstream,
    TYPE_ON_RESPONSE: on_response,
    TYPE_ON_RESPONSE_UPSTREAM: on_response_upstream,
    TYPE_ENCODER: encoder,
    TYPE_DECODER: decoder
}


class CheckerCategory:
    EDITOR = "Editor"
    CHECKER = "Checker"
    DEFAULT = "Other"

    @classmethod
    def get_order(cls, category):
        orders = [cls.EDITOR, cls.CHECKER, cls.DEFAULT]
        if category not in orders:
            return len(orders) + 1
        return orders.index(category)

    @classmethod
    def get_description(cls, category):
        descriptions = {
            cls.EDITOR: 'Modify flow', 
            cls.CHECKER: 'Check event and make notification', 
            cls.DEFAULT: 'Testability support, Advanced usage, etc.'
        }
        return descriptions.get(category, 'Custom category')

class LyrebirdCheckerServer(ThreadServer):
    def __init__(self):
        super().__init__()

        ROOT = Path(lyrebird.APPLICATION_CONF_DIR)
        self.SCRIPTS_DIR_TEMPLATE = ROOT/'checkers'
        self.EXAMPLE_DIR = Path(__file__).parent.parent/'examples'/'checkers'

        self.checkers = application.checkers
        self.load_checkers()

    def load_checkers(self):
        for checker_name in self.checkers:
            self.checkers[checker_name].update = False

        scripts_list = self.get_checker_list()
        if not scripts_list:
            return

        for checker_file in scripts_list:
            self.init_checker(checker_file.name, str(checker_file.absolute()))

        self.delete_unupdated_checker()

        logger.debug('------ checkers: ------')
        for name, checker in self.checkers.items():
            switch_conf = application.config.get('checker.switch', {})
            need_activate = switch_conf.get(name, False)
            self.activate_checker(checker, need_activate)

    def get_checker_list(self):
        workspace_str = application.config.get('checker.workspace')

        if workspace_str:
            workspace = Path(workspace_str)
            if not workspace.expanduser().exists():
                logger.error(f'Checker scripts dir {workspace_str} not found!')
                return
            workspace_iterdir = self.get_iterdir_python(workspace)
            if not workspace_iterdir:
                logger.warning(f'No checker script found in dir {workspace_str}')
                return

        else:
            workspace = Path(self.SCRIPTS_DIR_TEMPLATE)
            workspace.mkdir(parents=True, exist_ok=True)
            workspace_iterdir = self.get_iterdir_python(workspace)
            if not workspace_iterdir:
                self.copy_example_scripts()

        return workspace_iterdir

    @staticmethod
    def get_iterdir_python(path):
        path = Path(path)
        end_str = '.py'
        scripts_list = [i for i in path.iterdir() if i.suffix == end_str]
        return scripts_list

    def copy_example_scripts(self):
        for example in self.EXAMPLE_DIR.iterdir():
            if not example.name.endswith('.py'):
                continue
            dst = self.SCRIPTS_DIR_TEMPLATE / example.name
            shutil.copy(example, dst)

    def init_checker(self, name, path):
        checker = Checker(name, path)
        self.checkers[name] = checker

    def delete_unupdated_checker(self):
        del_name_list = []
        for name, checker in self.checkers.items():
            if not checker.update:
                del_name_list.append(name)

        for name in del_name_list:
            del self.checkers[name]

    def activate_checker(self, checker, need_activate):
        if need_activate and not checker.activated:
            checker.activate()
        elif not need_activate and checker.activated:
            checker.deactivate()
        logger.debug(checker.json())

    def load_scripts(self, scripts):
        scripts_list = [Path(s).expanduser() for s in scripts]
        for checker_file in scripts_list:
            self.init_checker(checker_file.name, str(checker_file.absolute()))
            self.activate_checker(self.checkers[checker_file.name], True)
            self.checkers[checker_file.name].debug = True
    
    def get_sorted_checker_groups(self):
        activated_checkers = {}
        deactivated_checkers = {}
        sorted_checker_groups = []
        for checker in self.checkers.values():
            category = checker.category
            activated = checker.activated
            checker_group = activated_checkers if activated else deactivated_checkers
            if category not in checker_group:
                checker_group[category] = []
            checker_group[category].append(checker.json())
        for checker_scripts in activated_checkers.values():
            checker_scripts.sort(key=lambda k: k['name'])
    
        for checker_scripts in deactivated_checkers.values():
            checker_scripts.sort(key=lambda k: k['name'])
    
        if activated_checkers:
            activated_checkers_list = [{
                'category': checker_category,
                'description': CheckerCategory.get_description(checker_category),
                'scripts': checker_scripts
            } for checker_category, checker_scripts in activated_checkers.items()]
            activated_checkers_list.sort(key=lambda k: CheckerCategory.get_order(k['category']))
            sorted_checker_groups.append({
                'status': 'Activated',
                'key': 'activated',
                'script_group': activated_checkers_list
            })
    
        if deactivated_checkers:
            deactivated_checkers_list = [{
                'category': checker_category,
                'description': CheckerCategory.get_description(checker_category),
                'scripts': checker_scripts
            } for checker_category, checker_scripts in deactivated_checkers.items()]
            deactivated_checkers_list.sort(key=lambda k: CheckerCategory.get_order(k['category']))
            sorted_checker_groups.append(
                {
                    'status': 'Deactivated',
                    'key': 'deactivated',
                    'script_group': deactivated_checkers_list
                }
            )
        return sorted_checker_groups

    def run(self):
        super().run()

    def stop(self):
        super().stop()


class Checker:
    def __init__(self, name, path):
        self.name = name
        self.activated = False
        self.path = path
        self.select = False
        self.debug = False
        self._module = None
        self._update = False
        self._funcs_map = {}
        self.title = '<No Title>'
        self.category = CheckerCategory.DEFAULT

        try:
            self._load_checker()
        except Exception as e:
            logger.error(f'Load checker {self.name} error:\n{e}\nfile path: {self.path}')

    @property
    def update(self):
        return self._update

    @update.setter
    def update(self, val):
        self._update = val

    def _load_checker(self):
        try:
            self._module = imp.load_source(self.name, self.path)
        except Exception as e:
            self.update = False
            raise CheckerIllegal(e)

        self.update = True
        self._checker_detection()
        global scripts_tmp_storage
        for func_type, func_list in scripts_tmp_storage.items():
            self._funcs_map[func_type] = [f for f in func_list]
        scripts_tmp_storage = {}

    def _checker_detection(self):
        if hasattr(self._module, TYPE_ENCODER) and not hasattr(self._module, TYPE_DECODER):
            logger.warning(f'Load checker {self.name} error: Encoder contains, but decoder not found!')

        if hasattr(self._module, TYPE_DECODER) and not hasattr(self._module, TYPE_ENCODER):
            logger.warning(f'Load checker {self.name} error: Decoder contains, but encoder not found!')
        
        if hasattr(self._module, 'TITLE'):
            self.title = self._module.TITLE
        
        if hasattr(self._module, 'CATEGORY'):
            self.category = self._module.CATEGORY

    def activate(self):
        for func_type, func_list in self._funcs_map.items():
            handler = FUNC_MAP_HANDLERS.get(func_type)
            if not handler:
                logger.error('Unhandler this func: ' + func_type)
                continue

            for func_info in func_list:
                handler.register(func_info)

        self.activated = True

    def deactivate(self):
        for func_type, func_list in self._funcs_map.items():
            handler = FUNC_MAP_HANDLERS.get(func_type)
            if not handler:
                logger.error('Unhandler this func: ' + func_type)
                continue

            for func_info in func_list:
                handler.unregister(func_info)

        self.activated = False

    def read(self):
        content = ''
        with codecs.open(self.path, 'r', 'utf-8') as f:
            content = f.read()
        return content

    def write(self, data):
        origin_activate_status = self.activated
        if origin_activate_status:
            self.deactivate()

        with codecs.open(self.path, 'w', 'utf-8') as f:
            f.write(data)
        self._load_checker()

        if origin_activate_status:
            self.activate()

    def json(self):
        return {k: self.__dict__[k] for k in self.__dict__ if not k.startswith('_')}

class CheckerIllegal(Exception):
    pass





class Plugin:

    def __init__(self, project_name, entry_point_name=None, version=None, location=None, **kwargs):
        self.project_name = project_name
        self.entry_point_name = entry_point_name
        self.version = version
        self.location = location
        self.manifest = ManifestProxy(**kwargs)
        self.status = []

    @staticmethod
    def default():
        return dict(
            view=('dist', 'index.html'),
            icon='mdi-widgets',
            api=[],
            background=[],
            event=[],
            on_request=[],
            on_response=[],
            on_request_upstream=[],
            on_response_upstream=[],
            status=[]
        )


class ManifestProxy:

    def __init__(self, **kwargs):
        self._manifest = kwargs
        self._set_default()

    def __getattr__(self, key):
        return self._manifest[key]

    def _set_default(self):
        default = Plugin.default()
        for default_key in default:
            if default_key in self._manifest:
                continue
            else:
                self._manifest[default_key] = default[default_key]



import re
import os
import math
import time
import socket
import tarfile
import requests
from pathlib import Path
from contextlib import closing
from lyrebird.log import get_logger

logger = get_logger()

def convert_size(size_bytes):
    if size_bytes == 0:
        return "0B"
    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    i = int(math.floor(math.log(size_bytes, 1024)))
    p = math.pow(1024, i)
    s = round(size_bytes / p, 2)
    return "%s %s" % (s, size_name[i])


def convert_time(duration):
    if duration < 1:
        return str(round(duration * 1000)) + 'ms'
    else:
        return str(round(duration, 2)) + 's'


def timeit(method):
    def timed(*args, **kw):
        ts = time.time()
        result = method(*args, **kw)
        te = time.time()
        print(f'{method.__name__} execution time {(te-ts)*1000}')
        return result
    return timed


def is_port_in_use(port, host='127.0.0.1'):
    sock = None
    try:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(1)
        sock.connect((host, int(port)))
        return True
    except socket.error:
        return False
    finally:
        if sock:
            sock.close()


def is_target_match_patterns(pattern_list, target):
    if not pattern_list or not target:
        return False
    for pattern in pattern_list:
        if TargetMatch.is_match(target, pattern):
            return True
    return False


def find_free_port():
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(('', 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


def compress_tar(input_path, output_path, suffix=None):
    current_path = Path.cwd()
    input_path = Path(input_path).expanduser().absolute().resolve()
    output_path = Path(output_path).expanduser().absolute().resolve()
    output_file = f'{output_path}{suffix}' if suffix else output_path

    os.chdir(input_path)
    # If not chdir, the directory in the compressed file will start from the root directory
    tar = tarfile.open(output_file, 'w:gz')
    for root, dirs, files in os.walk(input_path):
        for f in files:
            tar.add(f, recursive=False)
    tar.close()
    os.chdir(current_path)
    return output_file


def decompress_tar(input_path, output_path=None):
    input_path = Path(input_path).expanduser().absolute().resolve()
    if not output_path:
        filename = input_path.stem if input_path.suffix else f'{input_path.name}-decompress'
        output_path = input_path.parent / filename

    tf = tarfile.open(str(input_path))
    tf.extractall(str(output_path))
    tf.close()
    return output_path


def download(link, input_path):
    resp = requests.get(link, stream=True)
    with open(input_path, 'wb') as f:
        for chunck in resp.iter_content():
            f.write(chunck)


class CaseInsensitiveDict(dict):

    def __init__(self, dict):
        for k,v in dict.items():
            self.__setitem__(k, v)

    def __setitem__(self, key, value):
        super(CaseInsensitiveDict, self).__setitem__(key.lower(), value)

    def __getitem__(self, key):
        return super(CaseInsensitiveDict, self).__getitem__(key.lower())
    
    def get(self, key, default=None):
        return super(CaseInsensitiveDict, self).get(key.lower(), default)

class TargetMatch:

    @staticmethod
    def is_match(target, pattern):
        if not TargetMatch._match_type(target, pattern):
            return False
        if type(target) == str:
            return TargetMatch._match_string(target, pattern)
        elif type(target) in [int, float]:
            return TargetMatch._match_numbers(target, pattern)
        elif type(target) == bool:
            return TargetMatch._match_boolean(target, pattern)
        elif type(target).__name__ == 'NoneType':
            return TargetMatch._match_null(target, pattern)
        else:
            logger.warning(f'Illegal match target type: {type(target)}')
            return False

    @staticmethod
    def _match_type(target, pattern):
        return isinstance(target, type(pattern))

    @staticmethod
    def _match_string(target, pattern):
        return True if re.search(pattern, target) else False

    @staticmethod
    def _match_numbers(target, pattern):
        return target == pattern

    @staticmethod
    def _match_boolean(target, pattern):
        return target == pattern

    @staticmethod
    def _match_null(target, pattern):
        return target == pattern



from flask_restful import Resource
from lyrebird import application


class StatusBar(Resource):

    def get(self, item_id=None):
        if item_id == None:
            return self._get_all_status_item()
        else:
            return self._get_menu_by_status_item_name(item_id)

    def _get_all_status_item(self):
        status_item_ranked = []
        status_item_unranked = []
        plugin_server = application.server['plugin']
        for plugin_name in plugin_server.plugins:
            plugin = plugin_server.plugins[plugin_name]
            for status_item in plugin.status:
                if status_item.rank:
                    status_item_ranked.append({
                        'id': status_item.name,
                        'rank': status_item.rank,
                        'text': status_item.get_text()
                    })
                    continue
                status_item_unranked.append({
                    'id': status_item.name,
                    'text': status_item.get_text()
                })
        status_item_ranked.sort(key=lambda x:x['rank'])
        all_status_item = status_item_ranked + status_item_unranked
        return application.make_ok_response(data=all_status_item)

    def _get_menu_by_status_item_name(self, name):
        plugin_server = application.server['plugin']
        for plugin_name in plugin_server.plugins:
            plugin = plugin_server.plugins[plugin_name]
            for status_item in plugin.status:
                if status_item.name == name:
                    menu_list = [menu_item.json() for menu_item in status_item.get_menu()]
                    return application.make_ok_response(data=menu_list)
        return application.make_fail_response(f'Status item not found, id={name}')
